#include "bspline.h"
#include <cmath>

BSpline::BSpline(
        ESpline typeSpline,
        int k,
        EPolygone typePolygone,
        EVecteurNodal typeVecteur,
        int nbPtsCtrlX,
        int nbPtsCtrlZ,
        float stepX,
        float stepY)
{
    // Simple vérification
    switch (typeSpline) {
    case ESpline::SPLINE1D:
        assert(k <= nbPtsCtrlX);
        break;
    case ESpline::SPLINE2D:
        assert(k <= nbPtsCtrlX && k <= nbPtsCtrlZ);
        break;
    }

    // Initialisation des valeurs
    _k = k;
    _stepX = stepX;
    _stepY = stepY;
    srand(time(NULL));

    // Génération du polygone de contrôle
    construirePolygone(typePolygone, nbPtsCtrlX, nbPtsCtrlZ);

    // Calcul du vecteur nodal de nbPtsCtrl valeurs
    construireVecteurNodal(typeVecteur);
    switch (typeSpline) {
    case ESpline::SPLINE1D:
        calculerSpline1D();
        break;
    case ESpline::SPLINE2D:
        calculerSpline2D();
        break;
    default:
        std::cerr << "Type de spline inconnu" << std::endl;
        break;
    }

}

void BSpline::calculerSpline1D()
{
    // Init
    int dec = 0, m = _k - 1;
    uint i = 0;
    double u = 0;
    std::vector<glm::vec3> vecteursPointsControle (_nbPtsCtrlX);

    // Parcours de u
    for(u = _vecteurNodal[m]; u <= _vecteurNodal[_nbPtsCtrlX]; u += _stepX){
        dec = 0;
        // Calcul du décalage
        for(int i = _k; u > _vecteurNodal[i]; ++i)
            ++dec;

        // Récupération des k points de contrôles nécessaires à la floraison
        for(int i = 0; i < _k; i++){
            vecteursPointsControle[i] = _pointsDeControle[dec + i];
        }

        // Appel de la floraison récursive
        glm::vec3 newVertex = floraison(u, dec, _k, vecteursPointsControle);

        // Sauvegardes des paramètres de la B-spline
        _vertices.insert(_vertices.end(), {newVertex.x, newVertex.y, newVertex.z});
        _normals.insert(_normals.end(), {1, -1, -1});
        _indices.insert(_indices.end(), {i, i, ++i});
    }

    // Enlever la ligne d'indices en trop
    for(int i = 0; i < 3; ++i)
        _indices.pop_back();

    // Ajouter le polygone de controle aux vecteurs
    for(unsigned int i = 0; i < _pointsDeControle.size(); ++i) {
        _indices.insert(_indices.end(), {static_cast<unsigned int>(_vertices.size())/3, static_cast<unsigned int>(_vertices.size()/3), static_cast<unsigned int>(_vertices.size())/3+1});
        _vertices.insert(_vertices.end(), {_pointsDeControle[i].x, _pointsDeControle[i].y, _pointsDeControle[i].z});
        _normals.insert(_normals.end(), {0, 0, 0});
    }

    // Enlever la ligne d'indices en trop
    for(int i = 0; i < 3; ++i)
        _indices.pop_back();
}


void BSpline::calculerSpline2D()
{
}

glm::vec3 BSpline::floraison(float u, int dec, int k, std::vector<glm::vec3> vecteursPointsControle)
{
    if(k <= 1)
        return vecteursPointsControle.front();

    for(int i = 0; i < k - 1; ++i) {
        float max = _vecteurNodal[dec + k + i];
        float min = _vecteurNodal[dec + 1 + i];
        vecteursPointsControle[i] =
                ((( max - u) / (max - min)) * vecteursPointsControle[i]) +
                ((( u - min) / (max - min)) * vecteursPointsControle[i+1]);
    }

    return floraison(u, dec+1, k-1, vecteursPointsControle);
}

void BSpline::construirePolygone(EPolygone typePolygone, int nbPtsCtrlX, int nbPtsCtrlZ)
{
    // Intervalle pour afficher dans le repère caméra
    float xMin = -3, xMax =  3;
    float yMin = -2, yMax =  2;
    float zMin = -3, zMax = -6;

    switch (typePolygone) {
    case EPolygone::FIXE2D:
        _pointsDeControle.push_back(glm::vec3(-2, 0, -5));
        _pointsDeControle.push_back(glm::vec3(-1, 0, -5));
        _pointsDeControle.push_back(glm::vec3(-1, 1, -5));
        _pointsDeControle.push_back(glm::vec3( 1, 1, -5));
        _pointsDeControle.push_back(glm::vec3( 1, 0, -5));
        _pointsDeControle.push_back(glm::vec3( 2, 0, -5));
        _nbPtsCtrlX = 6;
        _nbPtsCtrlZ = 0;
        break;

    case EPolygone::RANDOM2D: {
        float x, y, z;
        for(int i = 0; i < nbPtsCtrlX; ++i) {
            //x = xMin + static_cast <float> (rand()) /( static_cast <float> (RAND_MAX/(xMax-xMin)));
            x = (i * xMin + (nbPtsCtrlX - i) * xMax) / (xMax - xMin);
            y = yMin + static_cast <float> (rand()) /( static_cast <float> (RAND_MAX/(yMax-yMin)));
            z = zMin + static_cast <float> (rand()) /( static_cast <float> (RAND_MAX/(zMax-zMin)));
            _pointsDeControle.push_back(glm::vec3(x, y, z));
        }
        _nbPtsCtrlX = _pointsDeControle.size();
        _nbPtsCtrlZ = 0;
        break;
    }

    case EPolygone::GAUSSIEN3D: {
        // Sigma = 1.0f
        float r, s = 2.0f;
        for(int x = - (nbPtsCtrlX / 2) + 1; x < nbPtsCtrlX / 2; ++x) {
            for(int z = - (nbPtsCtrlZ / 2) + 1; x < nbPtsCtrlZ / 2; ++x) {
                r = sqrt(x * x + z * z);
                _pointsDeControle.push_back(glm::vec3( x, (exp(-(r * r) / s)) / (M_PI * s), z));
            }
        }
        _nbPtsCtrlX = nbPtsCtrlX - 1 - (nbPtsCtrlX % 2);
        _nbPtsCtrlZ = nbPtsCtrlZ - 1 - (nbPtsCtrlZ % 2);
        break;
    }

    case EPolygone::FIXE3D: {
        float x, y, z;
        for(int i = 0; i < nbPtsCtrlX; ++i) {
            for(int j = 0; j < nbPtsCtrlZ; ++j) {
                x = (i * xMin + (nbPtsCtrlX - i) * xMax) / (xMax - xMin);
                y = fmod(x + z, 3);
                z = (j * zMin + (nbPtsCtrlZ - j) * zMax) / (zMax - zMin);
                _pointsDeControle.push_back(glm::vec3(x, y, z));
            }
        }
        _nbPtsCtrlX = nbPtsCtrlX;
        _nbPtsCtrlZ = nbPtsCtrlZ;
        break;
    }

    case EPolygone::RANDOM3D: {
        float x, y, z;
        for(int i = 0; i < nbPtsCtrlX; ++i) {
            for(int j = 0; j < nbPtsCtrlZ; ++i) {
                x = (i * xMin + (nbPtsCtrlX - i) * xMax) / (xMax - xMin);
                y = yMin + static_cast <float> (rand()) /( static_cast <float> (RAND_MAX/(yMax-yMin)));
                z = (j * zMin + (nbPtsCtrlZ - j) * zMax) / (zMax - zMin);
                _pointsDeControle.push_back(glm::vec3(x, y, z));
            }
        }
        _nbPtsCtrlX = nbPtsCtrlX;
        _nbPtsCtrlZ = nbPtsCtrlZ;
        break;
    }

    default:
        std::cerr << "Type de polygone inconnu" << std::endl;
        break;
    }
}

void BSpline::construireVecteurNodal(EVecteurNodal typeVecteur)
{
    switch (typeVecteur) {
    case EVecteurNodal::UNIFORME:
        for(int i = 0; i <= _k + _nbPtsCtrlX; ++i) {
            _vecteurNodal.push_back(i);
        }
        break;
    case EVecteurNodal::OUVERTUNIFORME:
        for(int i = 0; i <= _k + _nbPtsCtrlX; ++i) {
            if (i < _k)
                _vecteurNodal.push_back(0);
            else if (i > _nbPtsCtrlX)
                _vecteurNodal.push_back(_nbPtsCtrlX - _k + 1);
            else
                _vecteurNodal.push_back(i - _k + 1);
            std::cout << _vecteurNodal.back() << " ";
        }
        break;
    case EVecteurNodal::QUELCONQUE: {
        int r = 0;
        _vecteurNodal.push_back(0);
        for(int i = 1; i <= _k + _nbPtsCtrlX; ++i) {
            r = rand() / (RAND_MAX/5);
            _vecteurNodal.push_back(_vecteurNodal.back() + r);
        }
        break;
    }
    default:
        std::cerr << "Type de vecteur nodal inconnu" << std::endl;
        break;
    }

}
